#!/home/work/dts/php/bin/php
<?php
 
Class Analysis
{
    /**
     * config parameters
     */
    private $_config;
 
    /**
     * result
     */
    private $_results = array();
 
    /*
     * MySQL系统库名
     */
    private $_systemDatabase = array(
        'information_schema', 'mysql', 'performance_schema', 'sys'
    );
 
    /**
     * 支持的表字符集
     */
    private $_supportCharset = array(
        'utf8', 'gbk', 'latin1', 'binary', 'utf8mb4'
    );
    
    /*
     * 不支持的表字符集
     */
    private $_notSupportCharset = array(
        'utf16', 'utf16le', 'utf32', 'ucs2'
    );
 
    /*
     * 支持的数据类型
     *
     * boolean实际为tinyint(1)
     */
    private $_supportDataType = array(
        'bit', 'tinyint', 'boolean', 'smallint', 'mediumint', 'int', 'bigint',
        'decimal', 'float', 'double', 'date', 'datetime', 'timestamp', 'time',
        'year', 'char', 'varchar', 'binary', 'varbinary', 'enum', 'set',
        'tinytext', 'text', 'mediumtext', 'longtext', 'blob', 'tinyblob',
        'mediumblob', 'longblob',
        'json'
    );
 
    /*
     * 不支持的数据类型
     */
    private $_notSupportDataType = array(
        'geometry'
    );

    /*
     * 支持的存储引擎
     */
    private $_supportEngine = array(
        'InnoDB', 'MyISAM'
    );
 
    /**
     * 允许重试的错误码
     */
    private $_retryErrorNo = array(
        '2005', '2006', '2013', '2003', '1236'
    );
 
    /*
     * main function
     *
     * @param array    argv
     */
    public function main($argv)
    {
        if (count($argv) < 1) {
            printf("Usage: ./dts_analysis.php conf_file\n");
            return;
        }
 
        $config_file = $argv[1];
        if (!is_file($config_file)) {
            printf("%s is not exist", $config_file);
            return;            
        }
 
        //解析配置文件
        if (!$this->parseConfigFile($config_file)) {
            printf("parse config file error\n");
            return;
        }
 
        foreach ($this->_config["check_instance"] as $info) {
            //建立MySQL连接
            $conn = $this->initConnection($info);
            if (!$conn) {
                continue;
            }
 
            //检查MySQL版本
            if (!$this->checkVersion($info, $conn)) {
                continue;
            }
 
            //检查结构对象
            if (!$this->checkStructure($info, $conn)) {
                continue;
            }
 
            $allTables = $this->getAllTables($info, $conn);
            if (!$allTables || empty($allTables)) {
                continue;
            }
 
            //检查表字符集, 检查列字符集和数据类型
            $this->checkPerTable($info, $conn, $allTables);
        }
 
        //输出结果
        //$this->writeFile();
        foreach ($this->_results as $object) {
            if ($object["error"]) {
                printf("[ERROR] %s:%s: %s\n", $object["info"]["host"], $object["info"]["port"], $object["error"]);
            }
 
            if (!empty($object["unsupport"])) {
                foreach ($object["unsupport"] as $message) {
                    printf("[unsupport] %s:%s: %s\n", $object["info"]["host"], $object["info"]["port"], $message);
                }
            }
 
            if (!empty($object["risk"])) {
                foreach ($object["risk"] as $message) {
                    printf("[risk] %s:%s: %s\n", $object["info"]["host"], $object["info"]["port"], $message);
                }
            }
        }
 
        return;
    }
 
    /**
     * parse config file
     *
     * @param string    config file
     * @return          true if successfully, or false
     */
    public function parseConfigFile($conf_file)
    {
        $this->_config = json_decode(file_get_contents($conf_file), true);
        if (!isset($this->_config["check_instance"])) {
            printf("'check_instance' is not exists in config file\n");
            return false;
        }
 
        foreach ($this->_config["check_instance"] as $info) {
            if (!isset($info["host"]) || !isset($info["port"]) ||
                    !isset($info["user"]) || !isset($info["password"])) {
                return false;
            }
 
            //初始化结果集
            $index = $info["host"] . ":" . $info["port"];
            $newResult = array(
                "info" => $info,
                "error" => "",
                "unsupport" => array(),        
                "risk" => array(),
            );
            $this->_results[$index] = $newResult;
        }
 
        return true;
    }
 
    /**
     * init db connection
     *
     * @return  db connection object if init successfully, or false
     */
    public function initConnection($info)
    {
        $host = $info["host"];
        $port = $info["port"];
        $user = $info["user"];
        $pass = $info["password"];
        $dbname = "information_schema";
        $conn = null;
 
        for ($retry = 0; $retry <= 10; $retry ++) {
            try {
                $conn = mysqli_init();
                if (@mysqli_real_connect($conn, $host, $user, $pass, $dbname, $port)) {
                    return $conn;
                }
            } catch (Exception $e) {
                $conn = null;
            }
            $conn = null;
            sleep(1);
        }
 
        $index = $info["host"] . ":" . $info["port"];
        $this->_results[$index]["error"] = "Init connection failed";
        return false;
    }
 
    /*
     * 检查MySQL版本
     */ 
    public function checkVersion($info, $conn) {
        $index = $info["host"] . ":" . $info["port"];
 
        $sql = "select version() as version";
        $result = $this->executeSql($info, $conn, $sql);
        if (!$result) {
            return false;
        }
        $row = $result->fetch_array();
        $version = $row["version"];
        
        if (strstr($version, "secdb") || strstr($version, "risk")) {
            $this->_results[$index]["unsupport"][] = "[version]" . $version;
            return false;
        }
        
        list($first, $second) = explode(".", $version);
 
        if (!(intval($first) >= 5 && intval($second) >= 1)) {
            //小于5.1版本
            $this->_results[$index]["unsupport"][] = "[version]" . $version;
            return false;
        }
 
        $result->free();
        return true;
    }
 
    public function getAllTables($info, $conn) {
        $ret = array();
        $sql = "SELECT TABLE_SCHEMA,TABLE_NAME,TABLE_TYPE,ENGINE,TABLE_COLLATION,ROW_FORMAT FROM INFORMATION_SCHEMA.TABLES";
        $result = $this->executeSql($info, $conn, $sql);
        if (!$result) {
            return false;
        }
        
        while ($row = $result->fetch_array()) {
            //过滤系统库
            if (in_array($row["TABLE_SCHEMA"], $this->_systemDatabase)) {
                continue;
            }
            
            //记录"BASE TABLE"                
            if ("BASE TABLE" == $row["TABLE_TYPE"]) {
                $charset = substr($row["TABLE_COLLATION"], 0, strpos($row["TABLE_COLLATION"], "_"));
                $newObject = array(
                    "database" => $row["TABLE_SCHEMA"],
                    "table" => $row["TABLE_NAME"],
                    "charset" => $charset,
                    "row_format" => $row["ROW_FORMAT"],
                    "engine" => $row["ENGINE"]
                );
                $ret[] = $newObject;
            }                    
        }
 
        $result->free();
        return $ret;
    }
 
    /*
     * 校验结构对象
     *
     * 支持: database, table, view, index
     * 不支持: trigger(触发器), event(事件), tablespace, logfile group, server
     * 存在风险: reference(外键), procedure(存储过程), function(函数), partition(分区表)
     */ 
    public function checkStructure($info, $conn) {
        $index = $info["host"] . ":" . $info["port"];
 
        //判断是否存在trigger触发器
        $sql = "SELECT TRIGGER_SCHEMA,TRIGGER_NAME FROM INFORMATION_SCHEMA.TRIGGERS";
        $result = $this->executeSql($info, $conn, $sql);
        if (!$result) {
            return false;
        }
        while ($row = $result->fetch_array()) {
            if (in_array($row["TRIGGER_SCHEMA"], $this->_systemDatabase)) {
                continue;
            }
            $this->_results[$index]["unsupport"][] = 
                sprintf("[Trigger] db_name: %s, trigger_name: %s", $row["TRIGGER_SCHEMA"], $row["TRIGGER_NAME"]);
        }
        $result->free();
 
        //判断是否存在event(事件)
        $sql = "SELECT EVENT_SCHEMA,EVENT_NAME FROM INFORMATION_SCHEMA.EVENTS";
        $result = $this->executeSql($info, $conn, $sql);
        if (!$result) {
            return false;
        }
        while ($row = $result->fetch_array()) {
            if (in_array($row["EVENT_SCHEMA"], $this->_systemDatabase)) {
                continue;
            }
            $this->_results[$index]["unsupport"][] = 
                sprintf("[Event] db_name: %s, event_name: %s", $row["EVENT_SCHEMA"], $row["EVENT_NAME"]);
        }
        $result->free();
 
        //判断是否存在tablespace
        $sql = "SELECT TABLESPACE_NAME FROM INFORMATION_SCHEMA.TABLESPACES";
        $result = $this->executeSql($info, $conn, $sql);
        if (!$result) {
            return false;
        }
        while ($row = $result->fetch_array()) {
            $this->_results[$index]["unsupport"][] =
                sprintf("[Tablespace] tablespace_name: %s", $row["TABLESPACE_NAME"]);
        }
        $result->free();
 
        //判断是否存在logfile group
        $sql = "SELECT LOGFILE_GROUP_NAME FROM INFORMATION_SCHEMA.FILES WHERE LOGFILE_GROUP_NAME IS NOT NULL";
        $result = $this->executeSql($info, $conn, $sql);
        if (!$result) {
            return false;
        }
        while ($row = $result->fetch_array()) {
            $this->_results[$index]["unsupport"][] =
                sprintf("[logfileGroup] logfile_group_name: %s", $row["LOGFILE_GROUP_NAME"]);
        }
        $result->free();
 
        //判断是否存在server
        $sql = "SELECT Server_name FROM mysql.servers";
        $result = $this->executeSql($info, $conn, $sql);
        if (!$result) {
            return false;
        }
        while ($row = $result->fetch_array()) {
            $this->_results[$index]["unsupport"][] =
                sprintf("[Server] server_name: %s", $row["Server_name"]);
        }
        $result->free();
 
        //判断是否存在procedure(存储过程), function(函数)
        $sql = "SELECT ROUTINE_SCHEMA,ROUTINE_NAME,ROUTINE_TYPE FROM INFORMATION_SCHEMA.ROUTINES";
        $result = $this->executeSql($info, $conn, $sql);
        if (!$result) {
            return false;
        }
        while ($row = $result->fetch_array()) {
            if (in_array($row["ROUTINE_SCHEMA"], $this->_systemDatabase)) {
                continue;
            }
            if ($row["ROUTINE_TYPE"] == "PROCEDURE") {
                $this->_results[$index]["risk"][] =
                    sprintf("[Procedure] db_name: %s, procedure_name: %s", $row["ROUTINE_SCHEMA"], $row["ROUTINE_NAME"]);
                continue;
            }
            if ($row["ROUTINE_TYPE"] == "FUNCTION") {
                $this->_results[$index]["risk"][] =
                    sprintf("[Function] db_name: %s, function_name: %s", $row["ROUTINE_SCHEMA"], $row["ROUTINE_NAME"]);
            }
        }
        $result->free();
 
        //判断是否存在partition(分区表)
        $sql = "SELECT TABLE_SCHEMA,TABLE_NAME,PARTITION_NAME 
            FROM INFORMATION_SCHEMA.PARTITIONS WHERE PARTITION_NAME IS NOT NULL";
        $result = $this->executeSql($info, $conn, $sql);
        if (!$result) {
            return false;
        }
        while ($row = $result->fetch_array()) {
            if (in_array($row["TABLE_SCHEMA"], $this->_systemDatabase)) {        
                continue;
            }
            $this->_results[$index]["risk"][] =
                sprintf("[Partition] tb_name: %s.%s, partition_name: %s",
                    $row["TABLE_SCHEMA"], $row["TABLE_NAME"], $row["PARTITION_NAME"]);
        }
        $result->free();
 
        //判断是否存在含有reference外键的表
        $sql = "SELECT TABLE_SCHEMA,TABLE_NAME,COLUMN_NAME 
            FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE WHERE REFERENCED_COLUMN_NAME IS NOT NULL";
        $result = $this->executeSql($info, $conn, $sql);
        if (!$result) {
            return false;
        }
        while ($row = $result->fetch_array()) {
            if (in_array($row["TABLE_SCHEMA"], $this->_systemDatabase)) {
                continue;
            }
            $this->_results[$index]["risk"][] = 
                sprintf("[Reference] tb_name: %s.%s, column_name: %s",
                    $row["TABLE_SCHEMA"], $row["TABLE_NAME"], $row["COLUMN_NAME"]);
        }
        $result->free();
 
        return true;
    }
 
    /*
     * 检查表字符集, 检查列字符集, 检查列数据类型
     */ 
    public function checkPerTable($info, $conn, $allTables) {
        $index = $info["host"] . ":" . $info["port"];
 
        foreach ($allTables as $tbl) {
            //明确不支持的字符集
            if (in_array($tbl["charset"], $this->_notSupportCharset)) {
                $this->_results[$index]["unsupport"][] = 
                    sprintf("[UnsupportTableCharset] tb_name: %s.%s, tb_charset: %s",
                        $tbl["database"], $tbl["table"], $tbl["charset"]);
                continue;
            }
 
            //未知字符集
            if (!in_array($tbl["charset"], $this->_supportCharset)) {
                $this->_results[$index]["risk"][] =
                    sprintf("[UnknownTableCharset] tb_name: %s.%s, tb_charset: %s",
                        $tbl["database"], $tbl["table"], $tbl["charset"]);
            }
 
            $sql = sprintf("SELECT COLUMN_NAME,DATA_TYPE,COLUMN_DEFAULT,CHARACTER_SET_NAME 
                FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA='%s' AND TABLE_NAME='%s'",
                $tbl["database"], $tbl["table"]);
            $result = $this->executeSql($info, $conn, $sql);
            if (!$result) {
                return false;
            }
 
            while ($row = $result->fetch_array()) {
                // 有风险的存储引擎
                if (!in_array($tbl["engine"], $this->_supportEngine)) {
                    $this->_results[$index]["risk"][] =
                        sprintf("[UnknownTableEngine] tb_name: %s.%s, table_engine: %s",
                            $tbl["database"], $tbl["table"], $tbl["engine"]);
                }
                
                //不支持的列数据类型
                if (in_array($row["DATA_TYPE"], $this->_notSupportDataType)) {
                    $this->_results[$index]["unsupport"][] =
                        sprintf("[UnsupportColumnDataType] tb_name: %s.%s, column_name: %s, data_type: %s",
                            $tbl["database"], $tbl["table"], $row["COLUMN_NAME"], $row["DATA_TYPE"]);
                }

                //列字符集与表字符集不一致
                if ($row["CHARACTER_SET_NAME"] && ($row["CHARACTER_SET_NAME"] !== $tbl["charset"])) {
                    if ($row["CHARACTER_SET_NAME"] != "utf8" || $tbl["charset"] != "utf8mb4") {
                        $this->_results[$index]["risk"][] =
                            sprintf("[MismatchColumnCharset] tb_name: %s.%s, ".
                                "tb_charset: %s, column_name: %s, column_charset: %s",
                                $tbl["database"], $tbl["table"], $tbl["charset"],
                                $row["COLUMN_NAME"], $row["CHARACTER_SET_NAME"]);
                    }
                }
 
                //列数据类型为json
                if ("json" == $row["DATA_TYPE"] && $tbl["charset"] != "utf8mb4") {
                    $this->_results[$index]["risk"][] = 
                        sprintf("[JsonUtf8mb4] tb_name: %s.%s, tb_charset: %s, column_name: %s, data_type: %s",
                            $tbl["database"], $tbl["table"], $tbl["charset"],
                            $row["COLUMN_NAME"], $row["DATA_TYPE"]);
                }
 
                //列数据类型为datatime和timestamp, 且默认值为"0000-00-00 00:00:00"
                if ("datatime" == $row["DATA_TYPE"] || "timestamp" == $row["DATA_TYPE"]) {
                    if ("0000-00-00 00:00:00" == $row["COLUMN_DEFAULT"]) {
                        $this->_results[$index]["risk"][] = 
                            sprintf("[TimeDefaultZero] tb_name: %s.%s, column_name: %s, data_type: %s, default_value: %s",
                                $tbl["database"], $tbl["table"], $row["COLUMN_NAME"],
                                $row["DATA_TYPE"], $row["COLUMN_DEFAULT"]);
                    }
                }
 
                //未知列数据类型
                if (!in_array($row["DATA_TYPE"], $this->_supportDataType)) {
                    $this->_results[$index]["risk"][] =
                        sprintf("[UnknownDataType] tb_name: %s.%s, column_name: %s, data_type: %s",
                        $tbl["database"], $tbl["table"], $row["COLUMN_NAME"], $row["DATA_TYPE"]);
                }
            }
            $result->free();
        }
 
        return true;
    }
 
    /**
     * 执行sql, 执行失败自动重试
     *
     * @param          数据库连接信息
     * @param          MySQL连接
     * @param  string  待执行待sql
     * @param  int     重试次数
     * @return object  数据库执行结果
     */
    public function executeSql($info, &$conn, $sql, $retryNum = 10)
    {
        $result = @$conn->query($sql);
        $curRetry = 0;
 
        while (!$result && $curRetry < $retryNum &&
                (in_array($conn->errno, $this->_retryErrorNo))) {
            sleep(1);
            if ($conn = $this->initConnection($info)) {
                $result = @$conn->query($sql);
            }
            $curRetry++;
        }
 
        //错误码不在允许重试的错误码范围内, 则不重试
        if (!$result && !in_array($conn->errno, $this->_retryErrorNo)) {
            $index = $info["host"] . ":" . $info["port"];
            $this->_results[$index]["error"] = "Execute '" . $sql . "' failed, error number is " . $dbConn->errno;
        }
        return $result;
    }
 
    public function writeFile() {
        $curr = 1;
        foreach ($this->_results as $object) {
            $file_name = sprintf("output-%s.txt", $curr);
            $file = fopen($file_name, "w");
            $curr++;
 
            $msg = sprintf("host: %s, port: %s\n", $object["info"]["host"], $object["info"]["port"]);
            fwrite($file, $msg);
 
            if ($object["error"]) {
                $msg = sprintf("[ERROR] %s\n", $object["error"]);
                fwrite($file, $msg);
            }
            
            if (!empty($object["unsupport"])) {        
                foreach ($object["unsupport"] as $message) {
                    $msg = sprintf("[NotSupport] %s\n", $message);
                    fwrite($file, $msg);
                }
            }
            
            if (!empty($object["risk"])) {
                foreach ($object["risk"] as $message) {
                    $msg = sprintf("[Risk] %s\n", $message);
                    fwrite($file, $msg);
                }
            }
        }
    }
}
 
$obj = new Analysis();
$obj->main($argv);
?>
